// Block.as
// John Maloney, August 2009
//
// A Block is a graphical object representing a program statement (command)
// or function (reporter). A stack is a sequence of command blocks, where
// the following command and any nested commands (e.g. within a loop) are
// children. Blocks come in a variety of shapes and usually have some
// combination of of label strings and arguments (also children).
//
// The Block class manages block shape, labels, arguments, laout, and
// block sequence. It also supports generation of the labels and argument
// sequence from a specification string (e.g. "move %n").

package Blocks {
	import flash.display.DisplayObject;
	import flash.display.DisplayObjectContainer;
	import flash.display.Sprite;
	import flash.filters.GlowFilter;
	import flash.geom.Point;
	import flash.text.TextField;
	import flash.text.TextFieldAutoSize;
	import flash.text.TextFormat;

public class Block extends Sprite {

	public var spec:String;
	public var type:String;
	public var color:int;
	public var op:int;
	public var args:Array = [];

	public var isHat:Boolean = false;
	public var isReporter:Boolean = false;
	public var isTerminal:Boolean = false;	//  blocks that end a stack like "stop" or "forever"
	public var hasSubstack1:Boolean = false;
	public var hasSubstack2:Boolean = false;

	public var nextBlock:Block;
	public var subStack1:Block;
	public var subStack2:Block;
	
	public var base:BlockShape;
	public var labelsAndArgs:Array = [];
	private var argTypes:Array = [];

	private var indentLeft:int = 4, indentRight:int = 3;
	private var indentTop:int = 2, indentBottom:int = 2;

	public function Block(spec:String, type:String = "", color:int = 0xD00000, op:int = 0) {
		var shapeArray:Array;

		this.spec = spec;
		this.type = type;
		this.color = color;
		this.op = op;

		if (type == "b") {
			base = new BlockShape(Shapes.BooleanShape, color);
			isReporter = true;
			indentLeft = 10;
			indentRight = 9;
		} else if (type == "r") {
			base = new BlockShape(Shapes.NumberShape, color);	
			isReporter = true;
			indentLeft = 6;
			indentRight = 4;
		} else if (type == "h") {
			base = new BlockShape(Shapes.HatShape, color, Shapes.HatHSplit);
			isHat = true;
			indentTop = 12;
			indentBottom = 5;
		} else if ((type == "c") || (type == "cf")) {
			shapeArray = Shapes.LoopShape;
			if (type == "cf") {
				shapeArray = Shapes.FinalLoopShape;
				isTerminal = true;
			}
			hasSubstack1 = true;
			base = new BlockShape(shapeArray, color, Shapes.LoopHSplit);
			base.setSubSplit1(Shapes.LoopSub1Split, Shapes.LoopSub1OpeningH);
			indentTop = 2;
			indentBottom = 5;
		} else if (type == "f") {
			base = new BlockShape(Shapes.FinalCmdShape, color, Shapes.CmdHSplit);
			indentTop = 2;
			indentBottom = 3;
			isTerminal = true;
		} else {
			base = new BlockShape(Shapes.CmdShape, color, Shapes.CmdHSplit);
			indentTop = 3;
			indentBottom = 5;
		}
		addChild(base);
		addLabelsAndArgs(spec, color);
	}

	private function addLabelsAndArgs(spec:String, c:int):void {
		var specParts:Array = parseSpec(spec), i:int;
		labelsAndArgs = [];
		argTypes = [];
		for (i = 0; i < specParts.length; i++) {
			var o:DisplayObject = argOrLabelFor(specParts[i], c);
			labelsAndArgs.push(o);
			argTypes.push((o is BlockArg) ? specParts[i] : "label");
			addChild(o);
		}
		fixArgLayout();
	}

	public function setDefautArgs(defaultArgs:Array):void {
		var i:int, j:int = 0;
		for (i = 0; i < labelsAndArgs.length; i++) {
			if (labelsAndArgs[i] is BlockArg) {
				if (j < defaultArgs.length) {
					BlockArg(labelsAndArgs[i]).setDefault(defaultArgs[j]);
					j++;
				}
			}
		}
		fixArgLayout();
	}

	public function showRunFeedback():void {
		var f:GlowFilter = new GlowFilter(0xE000);
		f.strength = 3;
		f.blurX = f.blurY = 12;
		f.quality = 3;
		filters = [f];
	}

	public function hideRunFeedback():void {
		filters = [];
	}

	public function fixExpressionLayout():void {
		// fix expression layout up to the enclosing command block
		var b:Block = this;
		while (b.isReporter) {
			b.fixArgLayout();
			if (b.parent is Block) b = Block(b.parent)
			else return;
		}
		if (b is Block) b.fixArgLayout();
	}

	public function fixArgLayout():void {
		var i:int, x:int = indentLeft, maxH:int = 0;
		var item:DisplayObject;

		for (i = 0; i < labelsAndArgs.length; i++) {
			item = labelsAndArgs[i];
			item.x = x;
			maxH = Math.max(maxH, item.height);
			x += item.width + 2;
		}

		for (i = 0; i < labelsAndArgs.length; i++) {
			item = labelsAndArgs[i];
			item.y = indentTop + ((maxH - item.height) / 2);
		}
		base.setWidthHeight(x + indentRight, indentTop + maxH + indentBottom);
		if (type == "c") fixStackLayout();
		collectArgs();
	}

	public function setDefaultArgs(defaults:Array):void {
		var i:int;
		for (i = 0; i < Math.min(args.length, defaults.length); i++) {
			var v:* = defaults[i];
			if (v is BlockArg) v = BlockArg(v).argValue;
			if (args[i] is BlockArg) args[i].setDefault(v);
		}
	}

	private function collectArgs():void {
		var i:int;
		args = [];
		for (i = 0; i < labelsAndArgs.length; i++) {
			var a:* = labelsAndArgs[i];
			if ((a is Block) || (a is BlockArg)) args.push(a);
		}
	}

	public function fixStackLayout():void {
		var b:Block = this;
		while (b != null) {
			if (b.base.hasSubStack()) {
				var subH:int = Shapes.LoopSub1OpeningH;
				if (b.subStack1 != null) {
					b.subStack1.fixStackLayout();
					subH = b.subStack1.height;
					b.subStack1.x = 15;
					b.subStack1.y = b.base.substackY();
				}
				b.base.setSubstackHeight(subH);
			}
			if (b.nextBlock != null) {
				b.nextBlock.x = 0;
				b.nextBlock.y = b.base.height - Shapes.CmdNotchInset;
			}
			b = b.nextBlock;
		}
	}

	public function duplicate():Block {
		var dup:Block = new Block(spec, type, color, op);
		dup.isHat = isHat;
		dup.isReporter = isReporter;
		dup.isTerminal = isTerminal;
		dup.copyArgs(args);
		if (nextBlock != null) dup.addChild(dup.nextBlock = nextBlock.duplicate());
		if (subStack1 != null) dup.addChild(dup.subStack1 = subStack1.duplicate());
		if (subStack2 != null) dup.addChild(dup.subStack2 = subStack2.duplicate());
		dup.fixArgLayout();
		dup.fixStackLayout();
		return dup;
	}

	private function copyArgs(srcArgs:Array):void {
		// called on a newly created block that is being duplicated to copy the
		// argument values and/or expressions from the source block's arguments
		var i:int;
		collectArgs();
		for (i = 0; i < srcArgs.length; i++) {
			var argToCopy:* = srcArgs[i];
			if (argToCopy is BlockArg) BlockArg(args[i]).setDefault(BlockArg(argToCopy).argValue);
			if (argToCopy is Block) {
				var newArg:Block = Block(argToCopy).duplicate();
				var oldArg:* = args[i];
				labelsAndArgs[labelsAndArgs.indexOf(oldArg)] = newArg;
				args[i] = newArg;
				removeChild(oldArg);
				addChild(newArg);
			}
		}
	}

	public function removeBlock(b:Block):void {
		if (b.parent == this) removeChild(b);
		if (b == nextBlock) nextBlock = null;
		if (b == subStack1) subStack1 = null;
		if (b == subStack2) subStack2 = null;
		if (b.isReporter) {
			var i:int = labelsAndArgs.indexOf(b);
			if (i < 0) return;
			var newArg:DisplayObject = argOrLabelFor(argTypes[i], base.color);
			labelsAndArgs[i] = newArg;
			addChild(newArg);
			fixExpressionLayout();
		}
		topBlock().fixStackLayout();
	}

	public function appendBlock(b:Block):void {
		var bottom:Block = bottomBlock();
		bottom.addChild(b);
		bottom.nextBlock = b;
	}

	public function insertBlock(b:Block):void {
		var oldNext:Block = nextBlock;

		if (oldNext != null) removeChild(oldNext);

		addChild(b);
		nextBlock = b;
		if (oldNext != null) b.appendBlock(oldNext);
		topBlock().fixStackLayout();
	}

	public function insertBlockAbove(b:Block):void {
		if (b.nextBlock != null) b.removeChild(b.nextBlock);
		b.x = this.x;
		b.y = this.y - b.height + Shapes.CmdNotchInset;
		parent.addChild(b);
		b.insertBlock(this);
	}

	public function insertBlockSub1(b:Block):void {
		var old:Block = subStack1;
		if (old != null) old.parent.removeChild(old);

		addChild(b);
		subStack1 = b;
		if (old != null) b.appendBlock(old);
		topBlock().fixStackLayout();
	}

	public function insertBlockSub2(b:Block):void {
		var old:Block = subStack2;
		if (old != null) removeChild(old);

		addChild(b);
		subStack2 = b;
		if (old != null) b.appendBlock(old);
		topBlock().fixStackLayout();
	}

	public function replaceArgWithBlock(oldArg:DisplayObject, b:Block, pane:DisplayObjectContainer):void {
		var i:int = labelsAndArgs.indexOf(oldArg);
		if (i < 0) return;
		
		// remove the old argument
		removeChild(oldArg);
		labelsAndArgs[i] = b;
		addChild(b);
		fixExpressionLayout();

		if (oldArg is Block) {
			// leave old block in pane
			var o:Block = owningBlock();
			var p:Point = pane.globalToLocal(o.localToGlobal(new Point(o.width + 5, (o.height - oldArg.height) / 2)));
			oldArg.x = p.x;
			oldArg.y = p.y;
			pane.addChild(oldArg);
		}
		topBlock().fixStackLayout();
	}

	public function owningBlock():Block {
		var b:Block = this;
		while (true) {
			if (b.parent is Block) {
				b = Block(b.parent);
				if (!b.isReporter) return b; // owning command block
			} else {
				return b; // top-level reporter block
			}
		}
		return b; // never gets here
	}

	public function topBlock():Block {
		var result:DisplayObject = this;
		while (result.parent is Block) result = result.parent;
		return Block(result);
	}

	public function bottomBlock():Block {
		var result:Block = this;
		while (result.nextBlock!= null) result = result.nextBlock;
		return result;
	}

	private function argOrLabelFor(s:String, c:int):DisplayObject {
		if (s == "%b") return new BlockArg("b", c);
		if (s == "%n") return new BlockArg("n", c, true);
		if (s == "%n") return new BlockArg("s", c);
		if (s == "%s") return new BlockArg("s", c, true);
		if (s == "%v") return new BlockArg("s", c);
		return makeLabel(s);
	}

	private function makeLabel(label:String):TextField {
		var format:TextFormat = new TextFormat();
		format.color = 0xFFFFFF;
		format.font = "Verdana";
		format.size = 12;

		var text:TextField = new TextField();
		text.autoSize = TextFieldAutoSize.LEFT;
		text.selectable = false;
		text.background = false;
		text.text = label;
		text.setTextFormat(format);
		return text;
	}

	private function parseSpec(spec:String):Array {
		var i:int, j:int, len: int = spec.length;
		var result:Array = [];
		i = 0;
		while (i < len) {
			j = spec.indexOf("%", i);
			if (j >= 0) {
				if (j > i) result.push(stripWhiteSpace(spec.slice(i, j)));
				if ((j + 1) < len) {
					result.push(stripWhiteSpace(spec.slice(j, j + 2)));
				} else {
					result.push("%");
				}
				i = j + 2;
			} else {
				result.push(stripWhiteSpace(spec.slice(i, len)));
				i = len;
			}
		}
		return result;
	}

	private function stripWhiteSpace(s:String):String {
		var i:int = 0, j:int = s.length;
		while (((i + 1) < s.length) && (s.charCodeAt(i) <= 32)) i++;
		while (((j - 1) >= 0) && (s.charCodeAt(j - 1) <= 32)) j--;
		return s.slice(i, j);
	}

	public function blockDescription():String {
		var s:String = "", i:int;
		for (i = 0; i < labelsAndArgs.length; i++) {
			if (labelsAndArgs[i] is TextField) {
				s += " " + TextField(labelsAndArgs[i]).text;
			}
			if (labelsAndArgs[i] is BlockArg) {
				s += " " + BlockArg(labelsAndArgs[i]).argValue;
			}
		}
		return s;
	}

}}
