// BlockShape.as
// John Maloney, August 2009
//
// BlockShape handles drawing and resizing of a block shape based on
// a path defined by a sequence of simple drawing commands. Every path
// starts with an "m" (moveTo) command. The path will automatically
// be closed by drawing a line from the final pen position back to
// its initial position.
//
// The 45 degree commands are used to make end caps that scale
// in both x and y when the block gets taller.
//
// Path commands:
//		"m" x y		-- moveTo
//		"l" x y		-- lineTo
//		"c" x y [r]	-- curveTo (convex) with optional roundness parameter
//		"u" x y	[r]	-- curveTo (concave) with optional roundness parameter
//		"l45" x y	-- 45 degree lineTo
//		"c45" x y 	-- 45 degree curveTo (convex)

package Blocks {
	import flash.display.Shape;
	import flash.filters.BevelFilter;
	import flash.filters.GlowFilter;

public class BlockShape extends Shape {
	public var color:uint;

	private var origPath:Array, scaledPath:Array;
	private var baseW:int, baseH:int, hSplit:int;
	private var sub1Split:int, sub1BaseH:int;
	private var has45DegreeSegs:Boolean;
	private var w:int, h:int;
	private var topExtra:int, sub1Extra:int;

	public function BlockShape(path:Array = null, blockColor:int = 0, hSplit:int = 0) {
		origPath = (path == null) ? Shapes.RectShape : path;
		color = blockColor;
		initGeometry();
		this.hSplit = (hSplit > 0) ? hSplit : baseW / 2;
		setWidthHeight(baseW, baseH);
		addBevelFilter();
	}

	public function setSubSplit1(sub1Split:int, sub1BaseH:int):void {
		baseH = this.sub1Split = sub1Split;
		this.sub1BaseH = sub1BaseH;
	}

	public function copyFeedbackShapeFrom(s:BlockShape, reporterFlag:Boolean):void {
		origPath = s.origPath;
		initGeometry();
		hSplit = s.hSplit;
		setWidthHeight(s.w, s.h);
		addGlowFilter(reporterFlag);
	}

	public function setColor(color:int):void {
		this.color = color;
		drawPath();
	}

	public function setWidthHeight(newW:int, newH:int):void {
		var i:int;

		if (newH < baseH) newH = baseH;
		if (has45DegreeSegs) newW = newW - (newH - baseH); // scaling 45 segs increases width
		if (newW < baseW) newW = baseW;

		scaledPath = origPath.slice();
		for (i = 0; i < origPath.length; i++) {
			scaleCmdAt(i, newW, newH);
		}
		normalizePath();
		drawPath();
		topExtra = newH - baseH;
	}

	public function setSubstackHeight(substackH:int):void {
		// assume setWidthHeight() has been run
		var i:int;

		if (sub1Split == 0) return;  // not a control structure ("c") block
		sub1Extra = Math.max(Shapes.LoopEmptyOpening, (topExtra + substackH) - sub1BaseH);
		for (i = 0; i < origPath.length; i++) {
			var origCmd:Array = origPath[i];
			if  (origCmd[2] > sub1Split) scaledPath[i][2] = origCmd[2] + sub1Extra;
		}
		normalizePath();
		drawPath();
	}

	public function hasSubStack():Boolean {
		return sub1Split > 0;
	}

	public function substackY():int {
		return baseH + topExtra - Shapes.CmdNotchInset;
	}

	private function initGeometry():void {
		baseW = baseH = 0;
		has45DegreeSegs = false;
		for (var i:int = 0; i < origPath.length; i++) {
			var cmd:Array = origPath[i];;
			if (cmd[1] > baseW) baseW = cmd[1];
			if (cmd[2] > baseH) baseH = cmd[2];
			if ((cmd[0] == "c45") || (cmd[0] == "l45")) has45DegreeSegs = true;
		}
		hSplit = baseW / 2;
	}
	
	private function addBevelFilter():void {
		var f:BevelFilter = new BevelFilter(1);
		f.blurX = f.blurY = 2;
		f.highlightAlpha = 0.3;
		f.shadowAlpha = 0.6;
		this.filters = [f];
	}

	private function addGlowFilter(reporterFlag:Boolean):void {
		var f:GlowFilter;
		if (reporterFlag) {
			f = new GlowFilter(0xFFFFFF);
			f.strength = 5;
			f.blurX = f.blurY = 8;
			f.quality = 2;
		} else {
			f = new GlowFilter(0xE0E0E0);
			f.strength = 12;
			f.blurX = f.blurY = 5;
		}
		f.knockout = true;
		this.filters = [f];
	}

	private function scaleCmdAt(i:int, newW:int, newH:int):void {
		var origCmd:Array = origPath[i];
		var newCmd:Array = origCmd.slice();
		newCmd[1] = Math.round(scaleCoord(origCmd[1], baseW, newW, hSplit));
		newCmd[2] = Math.round(scaleCoord(origCmd[2], baseH, newH, baseH / 2));

		if ((origCmd[0] == "c45") || (origCmd[0] == "l45")) {
			// fix 45 degree endpoints
			var lastX:int = scaledPath[i-1][1];
			var lastY:int = scaledPath[i-1][2];
			var dx:int = newCmd[1] - lastX;
			var dy:int = newCmd[2] - lastY;
			if ((dx * dy) > 0) {
				newCmd[1] = scaledPath[i-1][1] + dy;
			} else {
				scaledPath[i-1][1] = newCmd[1] + dy;
			}
		}
		scaledPath[i] = newCmd;
	}

	private function scaleCoord(n:int, maxN:int, newMax:int, split:int):int {
		var mid:int = maxN / 2;
		if (newMax < maxN) return n;
		if (Math.abs(n - mid) < 1) return (newMax / 2) + (n - mid); // mid point
		return (n > split) ? newMax - (maxN - n) : n;
	}

	private function normalizePath():void {
		var i:int, cmd:Array;
		var minX:int, minY:int, maxX:int, maxY:int;
		minX = maxX = scaledPath[0][1];
		minY = maxY = scaledPath[0][2];
		for (i = 0; i < scaledPath.length; i++) {
			cmd = scaledPath[i];
			if (cmd[1] < minX) minX = cmd[1];
			if (cmd[2] < minY) minY = cmd[2];
			if (cmd[1] > maxX) maxX = cmd[1];
			if (cmd[2] > maxY) maxY = cmd[2];
		}
		for (i = 0; i < scaledPath.length; i++) {
			cmd = scaledPath[i];
			cmd[1] = cmd[1] - minX;
			cmd[2] = cmd[2] - minY;
		}
		w = maxX - minX;
		h = maxY - minY;
	}

	private function drawPath():void {
		var startX:int, startY:int;

		graphics.clear();
		graphics.beginFill(color);
		for (var i:int = 0; i < scaledPath.length; i++) {
			var cmd:Array = scaledPath[i];
			var thisX:int = cmd[1], thisY:int = cmd[2];
			var prevX:int, prevY:int;

			switch(cmd[0].charAt(0)) {
			case "c":
			case "u":
			 	 // curvature controls roundness; default of 0.42 approximates a quarter circle
				var curvature:Number = (cmd.length > 3) ? cmd[3] : 0.42;
				drawCurve(cmd, prevX, prevY, thisX, thisY, curvature);
				break;
			case "l":
				graphics.lineTo(thisX, thisY);
				break;
			case "m":
				graphics.moveTo(thisX, thisY);
				startX = thisX;
				startY = thisY;
				break;
			}
			prevX = thisX;
			prevY = thisY;
		}
		graphics.lineTo(startX, startY);
//		drawDebug(true, true);
	}

	private function drawDebug(showOutline:Boolean, showPoints:Boolean):void {
		if (showOutline) {
			graphics.beginFill(0, 0);
			graphics.lineStyle(0.1);
			graphics.drawRect(0, 0, w, h);
			graphics.lineStyle();
		}
		if (showPoints) {
			var i:int;
			for (i = 0; i < scaledPath.length; i++) {
				graphics.beginFill(0x2020A0);
				graphics.drawCircle(scaledPath[i][1], scaledPath[i][2], 2);
			}
		}
	}

	private function drawCurve(cmd:Array, p1x:int, p1y:int, p2x:int, p2y:int, roundness:Number):void {
		// compute control point by following an orthogal vector from the midpoint
		// of the line between p1 and p2 scaled by roundness * dist(p1, p2)

		if (cmd[0].charAt(0) == "u") roundness = -roundness;  // concave curve

		var midX:Number = (p1x + p2x) / 2.0;
		var midY:Number = (p1y + p2y) / 2.0;
		var cx:Number = midX + (roundness * (p2y - p1y));
		var cy:Number = midY - (roundness * (p2x - p1x));
		graphics.curveTo(cx, cy, p2x, p2y);
	}

}}
