// ScriptsPane.as
// John Maloney, August 2009
//
// A ScriptsPane is a working area that holds blocks and stacks. It supports the
// logic that highlights possible drop targets as a block is being dragged and
// decides what to do when the block is dropped.

package UI {
	import Blocks.Block;
	import Blocks.BlockArg;
	import Blocks.BlockShape;
	import flash.display.DisplayObject;
	import flash.display.Sprite;
	import flash.geom.Point;

public class ScriptsPane extends Sprite {

	private static var INSERT_NORMAL:int = 0;
	private static var INSERT_ABOVE:int = 1;
	private static var INSERT_SUB1:int = 2;
	private static var INSERT_SUB2:int = 3;

	private var possibleTargets:Array = [];
	private var nearestTarget:Array = [];
	private var feedbackShape:BlockShape;
	private var color:int = 0x808080;

	public function ScriptsPane(color:int) {
		this.color = color;
	}

	public function setWidthHeight(w:int, h:int):void {
		graphics.clear();
		graphics.beginFill(color);
		graphics.drawRect(0, 0, w, h);
		graphics.endFill();
	}

	public function prepareToDrag(b:Block):void {
		findTargetsFor(b);
		addFeedbackShape();
	}

	public function draggingDone():void {
		hideFeedbackShape();
		possibleTargets = [];
	}

	public function updateFeedbackFor(b:Block):void {
		nearestTarget = nearestTargetForBlockIn(b, possibleTargets);
		if (nearestTarget != null) {
			feedbackShape.x = nearestTarget[0].x;
			feedbackShape.y = nearestTarget[0].y;
			if (!b.isReporter) {
				feedbackShape.copyFeedbackShapeFrom(b.base, false);
				if (nearestTarget[2] == INSERT_ABOVE) {  // connecting command block to top of a stack
					feedbackShape.y = nearestTarget[0].y - b.height;
				}
			} else {
				var t:* = nearestTarget[1];
				if (t is Block) feedbackShape.copyFeedbackShapeFrom(Block(t).base, true);
				if (t is BlockArg) feedbackShape.copyFeedbackShapeFrom(BlockArg(t).base, true);
			}
		} else {
			hideFeedbackShape();
		}
	}

	public function blockDropped(b:Block):void {
		addChild(b);
		if (nearestTarget == null) return;
		if (b.isReporter) {
			Block(nearestTarget[1].parent).replaceArgWithBlock(nearestTarget[1], b, this);
		} else {
			var targetCmd:Block = nearestTarget[1];
			switch (nearestTarget[2]) {
			case INSERT_NORMAL:
				targetCmd.insertBlock(b);
				break;
			case INSERT_ABOVE:
				targetCmd.insertBlockAbove(b);
				break;
			case INSERT_SUB1:
				targetCmd.insertBlockSub1(b);
				break;
			case INSERT_SUB2:
				targetCmd.insertBlockSub2(b);
				break;
			}
		}
	}

	public function clearRunFeedback():void {
		allStacksDo(function(b:Block):void { b.hideRunFeedback() });
	}

	private function allStacksDo(f:Function):void {
		var i:int;
		for (i = 0; i < numChildren; i++) {
			var child:DisplayObject = getChildAt(i);
			if (child is Block)  f(Block(child));
		}
	}
	private function findTargetsFor(b:Block):void {
		possibleTargets = [];
		var i:int;
		for (i = 0; i < numChildren; i++) {
			var child:DisplayObject = getChildAt(i);
			if (child is Block) {
				var target:Block = Block(child);
				if (b.isReporter) {
					findReporterTargetsIn(target);
				} else {
					if (!target.isReporter) {
						if ((b.nextBlock == null) && !b.isTerminal && !target.isHat) {
							// b is a single, non-terminal command block and target
							// is not a hat so bottom of b can connect to top of target
							var p:Point = target.localToGlobal(new Point(0, 0));
							possibleTargets.push([p, target, INSERT_ABOVE]);
						}
						if (!b.isHat) findCommandTargetsIn(target, b.bottomBlock().isTerminal);
					}
				}
			}
		}
	}

	private function findCommandTargetsIn(stack:Block, endsWithTerminal:Boolean):void {
		var target:Block = stack;
		while (target != null) {
			var p:Point = target.localToGlobal(new Point(0, 0));
			if (!target.isTerminal && (!endsWithTerminal || (target.nextBlock == null))) {
				// insert stack after target block:
				// target block must not be a terminal
				// if stack does not end with a terminal, it can be inserted between blocks
				// otherwise, it can only inserted after the final block of the substack
				p = new Point(p.x, p.y + target.base.height - 3);
				possibleTargets.push([p, target, INSERT_NORMAL]);
			}
			if (target.hasSubstack1) {
				p = target.localToGlobal(new Point(16, target.base.substackY()));
				possibleTargets.push([p, target, INSERT_SUB1]);
			}
			if (target.subStack1 != null) findCommandTargetsIn(target.subStack1, endsWithTerminal);
			if (target.subStack2 != null) findCommandTargetsIn(target.subStack2, endsWithTerminal);
			target = target.nextBlock;
		}
	}

	private function findReporterTargetsIn(stack:Block):void {
		var b:Block = stack, i:int;
		while (b != null) {
			var targetType:String = null;
			for (i = 0; i < b.labelsAndArgs.length; i++) {
				var o:DisplayObject = b.labelsAndArgs[i];
				if ((o is Block) || (o is BlockArg)) {
					var p:Point = o.localToGlobal(new Point(0, 0));
					possibleTargets.push([p, o, INSERT_NORMAL]);
					if (o is Block) findReporterTargetsIn(Block(o));
				}
			}
			if (b.subStack1 != null) findReporterTargetsIn(b.subStack1);
			if (b.subStack2 != null) findReporterTargetsIn(b.subStack2);
			b = b.nextBlock;
		}
	}

	private function addFeedbackShape():void {
		if (feedbackShape == null) feedbackShape = new BlockShape();
		feedbackShape.setWidthHeight(10, 10);
		hideFeedbackShape();
		addChild(feedbackShape);
	}

	private function hideFeedbackShape():void {
		if (feedbackShape != null) feedbackShape.x = feedbackShape.y = -1000;
	}

	private function nearestTargetForBlockIn(b:Block, targets:Array):Array {
		var threshold:int = b.isReporter ? 15 : 25;
		var i:int, minDist:int = 100000;
		var nearest:Array;
		var bTopLeft:Point = new Point(b.x, b.y);
		var bBottomLeft:Point = new Point(b.x, b.y + b.height - 3);

		for (i = 0; i < targets.length; i++) {
			var item:Array = targets[i];
			var dist:Number = (item[2] == INSERT_ABOVE) ?
				Point.distance(bBottomLeft, item[0]) :
				Point.distance(bTopLeft, item[0]);
			if ((dist < minDist) && (dist < threshold) && dropCompatible(b, item[1])) {
				minDist = dist;
				nearest = item;
			}
		}
		return (minDist < threshold) ? nearest : null;
	}

	private function dropCompatible(droppedBlock:Block, target:DisplayObject):Boolean {
		if (!droppedBlock.isReporter) return true; // dropping a command block
		var dropType:String = droppedBlock.type;
		var targetType:String = (target is Block) ? Block(target).type : BlockArg(target).type;
		if (targetType == "b") return dropType == "b";
		return true;
	}

}}
