// BlocksEngineDemo.as
// John Maloney, September 2009
//
// This is the top-level sprite for a simple graphics programming system. It is responsible
// for creating the top-level UI, which in this case includes a blocks palette, cateogory
// selector, scripts pane, drawing pane, and output log pane. It is also responsible for
// updating the pane layout when the window changes size and handling the top-level mouse
// logic, including block drag-and-drop and running blocks and stacks.

package {
	import Blocks.Block;
	import UI.BlockPalette;
	import UI.CategorySelector;
	import UI.ScriptsPane;
	import flash.display.DisplayObject;
	import flash.display.Sprite;
	import flash.display.StageAlign;
	import flash.display.StageScaleMode;
	import flash.events.Event;
	import flash.events.MouseEvent;
	import flash.geom.Point;
	import flash.text.TextField;
	import flash.text.TextFieldType;
	import flash.text.TextFormat;
	import flash.utils.getTimer;
	
[SWF("backgroundColor"="#606060")]

public class BlocksEngineDemo extends Sprite {

	public var scriptsPane:ScriptsPane;
	public var categorySelector:CategorySelector;
	public var palette:BlockPalette;
	public var drawPane:DrawPane;
	public var logPane:TextField;
	public var interp:Interpreter;

	public var dragging:Boolean = false;
	public var targetBlock:Block;

	public function BlocksEngineDemo() {
		stage.scaleMode = StageScaleMode.NO_SCALE;
		stage.align = StageAlign.TOP_LEFT;
		stage.frameRate = 60;

		scriptsPane = new ScriptsPane(0x707070);
		addChild(scriptsPane);

		categorySelector = new CategorySelector(CmdSpecs.commands, CmdSpecs.categories);
		categorySelector.initCategories(CmdSpecs.categories.slice(1, 9));
		categorySelector.x = categorySelector.y = 5;
		addChild(categorySelector);

		palette = new BlockPalette(250, 500, 0xA0A0A0);
		palette.x = categorySelector.x;
		palette.y = categorySelector.y + categorySelector.height + 5;
		addChild(palette);
		categorySelector.blockPalette = palette;
		categorySelector.select(1);

		drawPane = new DrawPane(320, 240);
		drawPane.x = 600;
		drawPane.y = 5;
		addChild(drawPane);

		logPane = createLogPane();
		addChild(logPane);

		interp = new Interpreter(drawPane, scriptsPane, logPane);

		stage.addEventListener(MouseEvent.MOUSE_DOWN, mouseDown);
		stage.addEventListener(MouseEvent.MOUSE_MOVE, mouseMove);
		stage.addEventListener(MouseEvent.MOUSE_UP, mouseUp);
		stage.addEventListener(Event.RESIZE, fixStageLayout);
		stage.addEventListener(Event.ENTER_FRAME, step);

		fixStageLayout(null);
	}

	private function step(e:Event):void {
		interp.stepThreads();
	}

	private function createLogPane():TextField {
		var logField:TextField = new TextField();
		logField.background = true;
		logField.type = TextFieldType.INPUT;
		logField.defaultTextFormat = new TextFormat("Verdana", 14);
		logField.x = logField.y = 10;
		logField.text =
			"Welcome!\n\n" +
			"Drag blocks to middle area to make stacks.\n" +
			"Click on a block or stack to run.\n" +
			"Click on the background to stop all.\n" +
			"Shift-click on a block or stack to copy it.\n";
		return logField;
	}

	private function fixStageLayout(evt:Event):void {
		if (scriptsPane != null) {
			scriptsPane.setWidthHeight(stage.stageWidth, stage.stageHeight);
		}
		if ((categorySelector != null) && (palette != null)) {
			palette.setWidthHeight(
				categorySelector.width,
				stage.stageHeight - categorySelector.height - 15);
		}
		if (drawPane != null) {
			drawPane.x = stage.stageWidth - drawPane.paneW - 5;
			if (logPane != null) {
				logPane.x = drawPane.x;
				logPane.y = drawPane.y + drawPane.paneH + 5;
				logPane.width = drawPane.width;
				logPane.height = stage.stageHeight - (logPane.y + 5);
			}
		}
	}

	private function mouseDown(evt:MouseEvent):void {
		drawPane.mouseIsDown = true;
		dragging = false;
		targetBlock = null;

		var t:* = evt.target;
		var o:DisplayObject = ownerThatIsA(evt.target, Block);
		if ((o is Block) &&
			!((t is TextField) && (TextField(t).type == TextFieldType.INPUT))) {
				// target is a Block but not an editable arg
				targetBlock = Block(o);
		}
	}

	private function mouseMove(evt:MouseEvent):void {
		if (!dragging && (targetBlock != null)) {
			var b:Block = targetBlock;
			var p:Point = globalToLocal(b.localToGlobal(new Point(0, 0)));
			if ((b.parent is BlockPalette) || evt.shiftKey) {
				b = b.duplicate();
				targetBlock = b;
			}
			if (b.parent is Block) Block(b.parent).removeBlock(b);
			if (b.parent != null) b.parent.removeChild(b);
			scriptsPane.prepareToDrag(b);
			addChild(b);
			b.x = p.x;
			b.y = p.y;
			b.startDrag();
		}
		if (targetBlock != null) scriptsPane.updateFeedbackFor(targetBlock);
		dragging = true;
	}

	private function mouseUp(evt:MouseEvent):void {
		drawPane.mouseIsDown = false;
		if (targetBlock != null) {
			if (dragging) {
				targetBlock.stopDrag();
				if (palette.hitTestPoint(evt.stageX, evt.stageY)) {
					removeChild(targetBlock);
				} else {
					scriptsPane.blockDropped(targetBlock);
				}
			} else {
				interp.toggleThread(targetBlock.topBlock());				
			}
		}
		if (evt.target is ScriptsPane) interp.stopAllThreads();
		scriptsPane.draggingDone();
		dragging = false;
		targetBlock = null;
	}

	private static function ownerThatIsA(o:*, c:Class):DisplayObject {
		if (!(o is DisplayObject)) return null;
		var result:DisplayObject = DisplayObject(o);
		while (o != null) {
			if (o is c) return o;
			o = o.parent;
		}
		return null;
	}

}}
