// Interpreter.as
// John Maloney, August 2009
//
// Interpreter is a simple yet efficient interpreter for a blocks language.
//
// Interpreters may seem mysterious, but this one is quite straightforward. Since every
// block knows what block (if any) follows it in a sequence of blocks, the interpreter
// simply executes the current block, then asks that block for the next block. The heart
// of the interpreter is the evalCmd() function, a big case statement that uses the block's
// opcode to decide what to do. Control structures are handled by pushing the current state
// onto an array and starting execution with the first block of the substack. When the end
// of a substack is reached, the previous state is popped. If the substack was the body of
// a loop, control yields to the next thread. If it was in a conditional, execution continues
// with the block after the conditional. If there is no next block, and no state to pop, the
// thread terminates.
//
// To add a command to the interpreter, just add a new case to evalCmd(). Command blocks
// usually perform some operation and return null, while reporters must return a value.
// Control structures are a little more tricky, but if you look at some of the existing
// control structure commands you'll get a sense of what to do.

package {
	import Blocks.Block;
	import Blocks.BlockArg;
	import UI.ScriptsPane;
	import flash.media.Sound;
	import flash.text.TextField;
	import flash.utils.Dictionary;
	import flash.utils.getTimer;

public class Interpreter {

	private var threads:Array = [];
	private var activeThread:Thread;
	private var globals:Dictionary = new Dictionary();
	private var yield:Boolean;
	private var stopScript:Boolean;
	private var stopAll:Boolean;
	private var timerBase:int;

	private var drawPane:DrawPane;
	private var scriptsPane:ScriptsPane;
	private var logPane:TextField;

	public function Interpreter(drawPane:DrawPane, scriptsPane:ScriptsPane = null, logPane:TextField = null) {
		this.drawPane = drawPane;
		this.scriptsPane = scriptsPane;
		this.logPane = logPane;
		timerBase = getTimer();
	}

	public function toggleThread(b:Block):void {
		if (b.isReporter) {
			// click on reporter shows value in log
			appendToLog(evalCmd(b));
			return;
		}
		var i:int, newThreads:Array = [], wasRunning:Boolean = false;
		for (i = 0; i < threads.length; i++) {
			if (threads[i].topBlock == b) {
				wasRunning = true;
			} else {
				newThreads.push(threads[i]);
			}
		}
		threads = newThreads;
		if (wasRunning) {
			b.hideRunFeedback();
		} else {
			b.showRunFeedback();
			startThread(b);
		}
	}

	private function startThread(b:Block):void {
		activeThread = new Thread(b);
		threads.push(activeThread);
	}

	public function stopAllThreads():void {
		var i:int;
		stopAll = true;
		threads = [];
		if (scriptsPane != null) scriptsPane.clearRunFeedback();
	}

	public function stepThreads():void {
		frameCount++;
		if (threads.length == 0) return;
		var i:int, newThreads:Array = [];
		stopAll = false;
		for (i = 0; i < threads.length; i++) {
			activeThread = threads[i];
			stepActiveThread();
			if (stopAll) return;
			if ((activeThread.block != null) && (!stopScript)) {
				newThreads.push(activeThread);
			} else {
				activeThread.topBlock.hideRunFeedback();
			}
		}
		threads = newThreads;
	}

	private function stepActiveThread():void {
		if (activeThread.block == null) return;
		yield = stopScript = false;
		while (true) {
			evalCmd(activeThread.block);
			if (yield) return;
			activeThread.block = activeThread.block.nextBlock;
			while (activeThread.block == null) {  // end of command list
				if (activeThread.stack.length == 0) return;  // end of entire stack
				activeThread.popState();
				if (activeThread.isLoop) return;
				else activeThread.block = activeThread.block.nextBlock;
			}
		}
	}

	private function evalCmd(b:Block):* {
		var a0:*, msecs:int;
		switch(b.op) {
		case CmdSpecs.NOOP:
			return null;
		case CmdSpecs.FOREVER:
			startCmdList(b.subStack1, true);
			return null;
		case CmdSpecs.REPEAT:
			if (activeThread.firstTime) {
				activeThread.tmp = Math.max(int(Number(arg(b, 0))), 0); // repeat count
				activeThread.firstTime = false;
			}
			if (activeThread.tmp > 0) {
				activeThread.tmp--; // decrement count
				startCmdList(b.subStack1, true);
			} else {
				activeThread.firstTime = true;
			}
			return null;
		case CmdSpecs.IF:
			if (arg(b, 0)) startCmdList(b.subStack1, false);
			return null;
		case CmdSpecs.WAIT:
			if (activeThread.firstTime) {
				activeThread.tmp = getTimer(); // start time
				activeThread.firstTime = false;
			}
			msecs = getTimer() - activeThread.tmp;
			if ((msecs < 0) || (msecs >= 1000.0 * Number(arg(b, 0)))) {
				// time expired
				activeThread.tmp = 0;
				activeThread.firstTime = true;
			} else {
				yield = true;
			}
			return null;
		case CmdSpecs.STOP:
			yield = stopScript = true;
			return null;
		case CmdSpecs.STOP_ALL:
			stopAllThreads();
			yield = true;
			return null;
		case CmdSpecs.PLUS:
			return Number(arg(b, 0)) + Number(arg(b, 1));
		case CmdSpecs.MINUS:
			return Number(arg(b, 0)) - Number(arg(b, 1));
		case CmdSpecs.TIMES:
			return Number(arg(b, 0)) * Number(arg(b, 1));
		case CmdSpecs.DIVIDE:
			return Number(arg(b, 0)) / Number(arg(b, 1));
		case CmdSpecs.MOD:
			return Number(arg(b, 0)) % Number(arg(b, 1));
		case CmdSpecs.LESS:
			return Number(arg(b, 0)) < Number(arg(b, 1));
		case CmdSpecs.EQUAL:
			return arg(b, 0) == arg(b, 1);
		case CmdSpecs.GREATER:
			return Number(arg(b, 0)) > Number(arg(b, 1));
		case CmdSpecs.AND:
			return arg(b, 0) && arg(b, 1);
		case CmdSpecs.OR:
			return arg(b, 0) || arg(b, 1);
		case CmdSpecs.NOT:
			return !arg(b, 0);
		case CmdSpecs.RANDOM:
			return random(Number(arg(b, 0)), Number(arg(b, 1)));
		case CmdSpecs.ROUND:
			return Math.round(Number(arg(b, 0)));
		case CmdSpecs.JOIN:
			return String(arg(b, 0)) + arg(b, 1);
		case CmdSpecs.GET_VAR:
			a0 = arg(b, 0);
			if (!(a0 is String)) return null;
			if (globals[a0] == undefined) globals[a0] = 0;
			return globals[a0];
		case CmdSpecs.SET_VAR:
			a0 = arg(b, 0);
			if (!(a0 is String)) return null;
			globals[a0] = arg(b, 1);
			return null;
		case CmdSpecs.CHANGE_VAR:
			a0 = arg(b, 0);
			if (!(a0 is String)) return null;
			if (globals[a0] == undefined) globals[a0] = 0;
			globals[a0] = Number(globals[a0]) + Number(arg(b, 1));
			return null;
		case CmdSpecs.MOUSE_DOWN:
			return drawPane.mouseIsDown;
		case CmdSpecs.MOUSE_X:
			return drawPane.mouseX;
		case CmdSpecs.MOUSE_Y:
			return drawPane.mouseY;
		case CmdSpecs.TIMER:
			return getTimer() - timerBase;
		case CmdSpecs.CLEAR_TIMER:
			timerBase = getTimer();
			return null;
		case CmdSpecs.CLEAR_GRAPHICS:
			drawPane.clear();
			return null;
		case CmdSpecs.SET_RGB:
			drawPane.setRGB(Number(arg(b, 0)), Number(arg(b, 1)), Number(arg(b, 2)));
			return null;
		case CmdSpecs.SET_HUE:
			drawPane.setHue(Number(arg(b, 0)));
			return null;
		case CmdSpecs.CHANGE_HUE:
			drawPane.setHue(drawPane.hue + Number(arg(b, 0)));
			return null;
		case CmdSpecs.RECTANGLE:
			drawPane.rectangle(Number(arg(b, 0)), Number(arg(b, 1)), Number(arg(b, 2)), Number(arg(b, 3)));
			return null;
		case CmdSpecs.ELLIPSE:
			drawPane.ellipse(Number(arg(b, 0)), Number(arg(b, 1)), Number(arg(b, 2)), Number(arg(b, 3)));
			return null;
		case CmdSpecs.PRINT:
			appendToLog(arg(b, 0));
			return null;
		case CmdSpecs.PRINT_FRAMERATE:
			printFrameRate();
			return null;
		case CmdSpecs.CLEAR_TEXT:
			clearLog();
			return null;
		case CmdSpecs.PLAY_POP:
			playPop();
			return null;
		}
		return null;
	}

	private function arg(b:Block, i:int):* {
		var a:* = b.args[i];
		if (a == null) return null; // can happen if not enough arg fields in command spec
		return (a is BlockArg) ? BlockArg(a).argValue : evalCmd(Block(a));
	}

	private function startCmdList(b:Block, isLoop:Boolean = false):void {
		if (b == null) {
			if (isLoop) yield = true;
			return;
		}
		activeThread.isLoop = isLoop;
		activeThread.pushState();
		activeThread.initForBlock(b);
		evalCmd(activeThread.block);
	}


	[Embed(source="pop.mp3")]
	private var popSound:Class;

	private function playPop():void {
		var snd:Sound = new popSound() as Sound;
		snd.play();
	}

	private function random(n1:Number, n2:Number):Number {
		var low:Number = (n1 <= n2) ? n1 : n2;
		var hi:Number = (n1 <= n2) ? n2 : n1;
		if (low == hi) return low;
		var result:Number = (Math.random() * (hi - low)) + low;
		// if both low and hi are ints, truncate the result to an int
		if ((int(low) == low) && (int(hi) == hi)) result = int(result);
		return result
	}

	private function appendToLog(obj:*):void {
		if (logPane == null) return;
		logPane.appendText("" + obj + "\n");
		logPane.scrollV = logPane.maxScrollV - 1;
	}

	private function clearLog():void {
		if (logPane == null) return;
		logPane.text = "";
	}

	private var firstFrameTime:int;
	private var frameCount:int;

	private function printFrameRate():void {
		var now:int = getTimer();
		var msecs:int = now - firstFrameTime;
		if (msecs > 200) {
			var fps:Number = Math.round((1000 * frameCount) / msecs);
			appendToLog(fps + " fps (" + Math.round(msecs / frameCount) + " msecs/frame)");
			firstFrameTime = now;
			frameCount = 0;
		}
	}

}}
